# -*- coding: utf-8 -*-
# Description:
# OSP Ovaldi Daemon wrapper
#
# Authors:
# Hani Benhabiles <hani.benhabiles@greenbone.net>
#
# Copyright:
# Copyright (C) 2014 Greenbone Networks GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

""" OSP ovaldi class. """

import base64
import logging
import os
import shutil
import socket

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

from ospd.ospd import OSPDError, OSPDaemon
from ospd.misc import main as daemon_main
from ospd_ovaldi import __version__

import paramiko

logger = logging.getLogger(__name__)

OSPD_OVALDI_DESC = """
This scanner runs the Open Source OVAL scanner 'ovaldi' being installed on the
target systems. To do so, a SSH access is required. Note that the current
version does not support Windows or other systems without SSH access.

The OVAL Interpreter is a freely available reference implementation that
demonstrates the evaluation of OVAL Definitions. Based on a set of Definitions
the Interpreter collects system information, evaluates it, and generates a
detailed OVAL Results file. It has been developed to demonstrate the usability
of OVAL Definitions and to ensure correct syntax and adherence to the OVAL
Schemas by definition writers.

IMPORTANT: Please note that the OVAL Interpreter is not an enterprise scanning
tool; it is a simplistic, command-line interface that has the ability to execute
OVAL Content on an end system.

For more see the homepage of ovaldi at MITRE:
    https://oval.mitre.org/language/interpreter.html
"""

OSPD_OVALDI_PARAMS = \
    {'username':
     {'type': 'string',
      'name': 'SSH Username',
      'default': '',
      'description': 'The SSH username used to log into the target and to run'
                     ' the ovaldi tool installed on that target.',},
     'password':
     {'type': 'password',
      'name': 'SSH Password',
      'default': '',
      'description': 'The SSH password for the given username which is used to'
                     ' log into the target and to run the ovaldi tool installed'
                     ' on that target. This should not be a privileged user'
                     ' like "root", a regular privileged user account should be'
                     ' sufficient in most cases.'},
     'port':
     {'type': 'integer',
      'name': 'SSH Port',
      'default': 22,
      'description': 'The SSH port which to use for logging in with the given'
                     ' username/password. the ovaldi tool installed on that'
                     ' target.',},
     'definitions_file':
     {'type': 'ovaldef_file',
      'name': 'Oval Definitions',
      'default': '',
      'description': 'OVAL definitions is a XML object containing many single'
                     ' oval definition objects including also any required oval'
                     ' test and other objects. Content should be base64'
                     ' encoded.',},
     'ssh_timeout':
     {'type': 'integer',
      'name': 'SSH timeout',
      'default': 30,
      'description': 'Timeout when communicating with the target via SSH.',},}


def get_definition_result(def_id, results):
    """ Gets an oval definition's result value in results element from
    results xml file. """
    for child in results:
        if child.tag.endswith('system'):
            system = child
            break
    for child in system:
        if child.tag.endswith('definitions'):
            definitions = child
            break
    for child in definitions:
        if child.attrib.get('definition_id') == def_id:
            return child.attrib.get('result')
    return "Not found"

def check_ovaldi(ssh, sftp, schema_dir):
    """
    Check that ovaldi and the required files are installed correctly on the
    target.

    Return None if success, error string otherwise.
    """
    chan = ssh.get_transport().open_session()
    chan.exec_command("which ovaldi")
    status = chan.recv_exit_status()
    chan.close()
    if status != 0:
        return "ovaldi not found on target host."

    # Is oval schema directory present ?
    try:
        sftp.stat(schema_dir)
    except IOError:
        return "oval schema folder {0} not found.".format(schema_dir)
    return None

class OSPDOvaldi(OSPDaemon):

    """ Class for ospd-ovaldi daemon. """

    def __init__(self, certfile, keyfile, cafile):
        """ Initializes the ospd-ovaldi daemon's internal data. """
        super(OSPDOvaldi, self).__init__(certfile=certfile, keyfile=keyfile,
                                         cafile=cafile)

        self.server_version = __version__
        self.scanner_info['name'] = 'ovaldi'
        self.scanner_info['version'] = 'depends on the local installation at the target host'
        self.scanner_info['description'] = OSPD_OVALDI_DESC
        for name, param in OSPD_OVALDI_PARAMS.items():
            self.add_scanner_param(name, param)
        self.schema_dir = "/usr/share/ovaldi/xml"

    def check(self):
        return True

    def process_scan_params(self, params):
        try:
            params['definitions'] = base64.b64decode(
                params['definitions_file'])
        except TypeError:
            raise OSPDError(
                "Couldn't decode base64 definitions file", 'start_scan')
        del params['definitions_file']

        return params

    def finish_scan_with_err(self, scan_id, target, local_dir=None,
                             err="Unknown error"):
        """
        Add an error message to a scan and finish it. Cleanup the results
        dir if provided.
        """
        if local_dir and logger.getEffectiveLevel() > logging.DEBUG:
            shutil.rmtree(local_dir)
        logger.info(err)
        self.add_scan_error(scan_id, host=target, value=err)

    def exec_scan(self, scan_id, target):
        """ Starts the ovaldi scanner for scan_id scan. """
        options = self.get_scan_options(scan_id)
        local_dir = '/tmp/ovaldi-results-{0}-{1}'.format(scan_id, target)
        os.mkdir(local_dir)
        defs_path = '{0}/definitions.xml'.format(local_dir)

        # Write definitions to temporary file
        with open(defs_path, 'w') as defs_file:
            defs_file.write(options['definitions'])

        # Connect to target
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            port = int(options['port'])
            timeout = int(options['ssh_timeout'])
            ssh.connect(hostname=target, username=options['username'],
                        password=options['password'], timeout=timeout,
                        port=port)
        except (paramiko.ssh_exception.AuthenticationException,
                socket.error) as err:
            # Errors: No route to host, connection timeout, authentication
            # failure etc,.
            return self.finish_scan_with_err(scan_id, target, local_dir,
                                             str(err))

        # Can we SFTP to the target ?
        try:
            sftp = ssh.open_sftp()
        except paramiko.ssh_exception.SSHException:
            err = "Couldn't SFTP to the target host."
            ssh.close()
            return self.finish_scan_with_err(scan_id, target, local_dir, err)

        # Check for ovaldi on the target.
        err = check_ovaldi(ssh, sftp, self.schema_dir)
        if err:
            sftp.close()
            ssh.close()
            return self.finish_scan_with_err(scan_id, target, local_dir, err)

        # Create temp dir and copy definitions file.
        target_dir = "/tmp/ovaldi-remote-{0}-{1}".format(scan_id, target)
        try:
            sftp.mkdir(target_dir)
        except IOError as err:
            err = "Failed to mkdir {0} on target: {1}".format(target_dir, err)
            sftp.close()
            ssh.close()
            return self.finish_scan_with_err(scan_id, target, local_dir, err)
        target_defs_path = "{0}/definitions.xml".format(target_dir)
        sftp.put(defs_path, target_defs_path)

        # Run ovaldi
        results_path = "{0}/results.xml".format(target_dir)
        syschar_path = "{0}/oval_syschar.xml".format(target_dir)
        log_path = "{0}/ovaldi.log".format(target_dir)
        command = "ovaldi -m -s -r {0} -d {1} -o {2} -y {3} -a {4}"\
            .format(results_path, syschar_path, target_defs_path,
                    target_dir, self.schema_dir)
        logger.info("Running command: {0}".format(command))
        _, stdout, _ = ssh.exec_command(command)
        # Flush stdout buffer, to continue execution.
        stdout.readlines()
        # Get the results files from target.
        # results.xml
        try:
            local_results = "{0}/results.xml".format(local_dir)
            sftp.get(results_path, local_results)
            self.parse_results_xml(local_results, scan_id, target)
        except IOError as err:
            msg = "Couldn't get results.xml: {0}".format(err)
            logger.info(msg)
            self.add_scan_error(scan_id, host=target, value=msg)
        # oval_syschar.xml
        try:
            local_syschar = "{0}/oval_syschar.xml".format(local_dir)
            sftp.get(syschar_path, local_syschar)
            self.parse_oval_syschar_xml(local_syschar, scan_id, target)
        except IOError as err:
            msg = "Couldn't get oval_syschar.xml: {0}".format(err)
            logger.info(msg)
            self.add_scan_error(scan_id, host=target, value=msg)
        if int(options['debug_mode']):
            try:
                local_log = "{0}/ovaldi.log".format(local_dir)
                sftp.get(log_path, local_log)
                self.parse_ovaldi_log(local_log, scan_id, target)
            except IOError as err:
                msg = "Couldn't get ovaldi.log: {0}".format(err)
                logger.info(msg)
                self.add_scan_error(scan_id, host=target, value=msg)
        # Cleanup temporary directories and close connection.
        sftp.close()
        if logger.getEffectiveLevel() > logging.DEBUG:
            ssh.exec_command("rm -rf {0}".format(target_dir))
        else:
            logger.debug("{0} not removed.".format(target_dir))

        ssh.close()
        if logger.getEffectiveLevel() > logging.DEBUG:
            shutil.rmtree(local_dir)

    def parse_oval_syschar_xml(self, file_path, scan_id, target):
        """ Parses the content of oval_syschar.xml file to scan results """

        try:
            with open(file_path, 'r') as syschar_file:
                file_content = syschar_file.read()

            # Extract /oval_system_characteristcs/system_info
            system_info = None
            tree = ET.fromstring(file_content)
            for child in tree:
                if child.tag.endswith('system_info'):
                    system_info = child
                elif child.tag.endswith('generator'):
                    generator = child
            for child in generator:
                value = child.tag
                if '}' in child.tag:
                    value = child.tag.split('}')[1]
                name = 'syschar_generator:{0}'.format(value)
                self.add_scan_log(scan_id, host=target, name=name,
                                  value=child.text)
                if value == 'product_name' and child.text.startswith('cpe:'):
                    self.add_scan_host_detail(scan_id, host=target, name="App",
                                              value=child.text)

            for child in system_info:
                if not child.tag.endswith('interfaces'):
                    name = child.tag.split('}')[1]
                    self.add_scan_host_detail(scan_id, host=target, name=name,
                                              value=child.text)
                else:
                    # Extract interfaces info from <sytem_info><interfaces>...
                    self.parse_system_info_interfaces(child, scan_id, target)

        except IOError:
            logger.debug("{0}: Couldn't open file.".format(file_path))

    def parse_system_info_interfaces(self, interfaces, scan_id, target):
        """ Parses interfaces information in ovaldi's system_info's interfaces
        and insert it in scan results. """

        for interface in interfaces:
            name = str()
            address = str()
            mac = str()
            for child in interface:
                if child.tag.endswith('interface_name'):
                    name = child.text
                elif child.tag.endswith('ip_address'):
                    address = child.text
                elif child.tag.endswith('mac_address'):
                    mac = child.text
            result_str = '{0}|{1}|{2}'.format(name, address, mac)
            self.add_scan_host_detail(scan_id, host=target, value=result_str,
                                      name='interface')

    def parse_results_xml(self, file_path, scan_id, target):
        """ Parses results file into scan results. """

        try:
            with open(file_path, 'r') as results_file:
                file_content = results_file.read()
            # Extract oval definitions results and other relevant information.
            tree = ET.fromstring(file_content)
            for child in tree:
                if child.tag.endswith('generator'):
                    generator = child
                elif child.tag.endswith('oval_definitions'):
                    oval_defs = child
                elif child.tag.endswith('results'):
                    results = child
            for child in generator:
                value = child.tag
                if '}' in child.tag:
                    value = child.tag.split('}')[1]
                name = 'results_generator:{0}'.format(value)
                self.add_scan_log(scan_id, host=target, name=name,
                                  value=child.text)
            self.parse_oval_results(oval_defs, results, scan_id, target)
        except IOError:
            logger.debug("{0}: Couldn't open file.".format(file_path))

    def parse_oval_results(self, oval_defs, results, scan_id, target):
        """ Parses oval_definitions and results elements from results file. """

        for child in oval_defs:
            if child.tag.endswith('generator'):
                generator = child
            elif child.tag.endswith('definitions'):
                definitions = child
        for child in generator:
            name = 'defs_generator:{0}'.format(child.tag.split('}')[1])
            self.add_scan_log(
                scan_id, host=target, name=name, value=child.text)
        for definition in definitions:
            def_id = definition.attrib.get('id')
            def_result = get_definition_result(def_id, results)
            if def_result == 'true':
                # Skip: false, error, unknown, not applicable.
                self.add_scan_alarm(
                    scan_id, host=target, name=def_id, value='', test_id=def_id)

    def parse_ovaldi_log(self, file_path, scan_id, target):
        """ Parses the content of ovaldi.log file to scan results """

        try:
            with open(file_path, 'r') as log_file:
                file_content = log_file.read()
                self.add_scan_log(scan_id, host=target, name="ovaldi.log",
                                  value=file_content)
        except IOError:
            logger.debug("{0}: Couldn't open file.".format(file_path))


def main():
    """ OSP ovaldi main function. """
    daemon_main('OSPD - ovaldi wrapper', OSPDOvaldi)
